/* Copyright Izak Burger 2005.
 * Licensed under the GPL. */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>

#define PAM_SM_SESSION
#define PAM_SM_AUTH
/* #define PAM_SM_PASSWORD */

#define EXPOSE_AUTHTOK 1
#define EXPOSE_KRB5CCNAME 2

#include <security/_pam_macros.h>
#include <security/pam_modules.h>

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define LOCAL_LOG_PREFIX "PAM-script: "

struct pamscript_options {
	uid_t uid;
	gid_t gid;
	char cmd[256];
	char *realuser;
	char *service;
	int cmdoverride;
	int expose;
};

static void _pam_log(int err, const char *format, ...)
{
	va_list args;
	va_start(args, format);
	vsyslog(LOG_AUTH | err, format, args);
	va_end(args);
}

/* Stolen and adapted from pam_ldap */
static int _get_authtok (pam_handle_t *pamh){
	int rc;
	char *p;
	struct pam_message msg[1], *pmsg[1];
	struct pam_response *resp; 
	struct pam_conv *conv;

	pmsg[0] = &msg[0];
	msg[0].msg_style = PAM_PROMPT_ECHO_OFF;
	msg[0].msg = "Password: ";
	resp = NULL; 
	rc = pam_get_item(pamh, PAM_CONV, (const void **)&conv);
	if (rc == PAM_SUCCESS){
		rc = conv->conv (1,
				(const struct pam_message **) pmsg,
				&resp, conv->appdata_ptr);
	} else {
		return rc;
	}

	if (resp != NULL){
		p = resp[0].resp;
		/* leak if resp[0].resp is malloced. */
		resp[0].resp = NULL;
	} else {
		return PAM_CONV_ERR;
	}

	free (resp);
	pam_set_item (pamh, PAM_AUTHTOK, p);

	return PAM_SUCCESS;
}


static int parse_options(pam_handle_t *pamh,
		int argc, const char **argv,
		const char *execname, struct pamscript_options *options){
	char *user;
	char myuser[256];
	struct passwd *pw;
	int i;

	pam_get_item(pamh, PAM_SERVICE, (const void**)&(options->service));

	memset(options->cmd, '\0', sizeof(options->cmd));
	snprintf(options->cmd, sizeof(options->cmd), "/etc/security/%s", execname);
	options->cmdoverride = 0;
	options->expose = 0;

	/* Get the username of the person logging in right now. */
	pam_get_user(pamh, (const char**)&user, NULL);
	if(user == NULL){
		_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Cannot determine name of user");
		return 0;
	}
	options->realuser = user;

	/* Handle arguments */
	for (i=0; i < argc; i++) {
		char *junk;
		char mybuf[256];
		memset(mybuf,'\0',sizeof(mybuf));
		junk = strchr(argv[i], '=');
		/* If the option is too long or it contains no '=', it must be bad */
		if ((junk == NULL) || (junk - argv[i]) >= sizeof(mybuf)) {
			_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Bad option: \"%s\"",
			    argv[i]);
			continue; /* next argument */
		}
		strncpy(mybuf, argv[i], MIN(junk - argv[i], sizeof(mybuf) - 1));
		if (strcmp(mybuf, execname)==0) {
			memset(options->cmd, '\0', sizeof(options->cmd));
			strncpy(options->cmd, junk + 1, sizeof(options->cmd) - 1);
			options->cmdoverride = 1;
		} else if (strcmp(mybuf, "runas")==0) {
			strncpy(myuser, junk + 1, sizeof(myuser) - 1);
			user = myuser;
		} else if(strcmp(mybuf, "expose")==0) {
			options->expose = atoi(junk + 1);
		}
	}

	pw = getpwnam(user);
	/* Unable to get password info, return session error */
	if (pw==NULL){
		_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Unable to get passwd entry for %s", user);
		return 0;
	}

	options->uid = pw->pw_uid;
	options->gid = pw->pw_gid;

	return 1;
}

static int session_script(struct pamscript_options options, int numenv, ...)
{
	pid_t pid;
	sigset_t newmask, oldmask;
	int retcode = 0;

	_pam_log(LOG_DEBUG, LOCAL_LOG_PREFIX "Real User is: %s", options.realuser);
	_pam_log(LOG_DEBUG, LOCAL_LOG_PREFIX "Command is:   %s", options.cmd);
	_pam_log(LOG_DEBUG, LOCAL_LOG_PREFIX "Executing uid:gid is: %d:%d", options.uid, options.gid);

	/* Block SIGCHLD */
	sigemptyset(&newmask);
	sigaddset(&newmask, SIGCHLD);
	sigprocmask(SIG_BLOCK, &newmask, &oldmask);

	/* Time to fork, we check for existence and executability of the script later, because we
	 * need to do that as the user executing the script, as it is possible that the script is
	 * executable by root but not by the user */
	pid = fork();
	if(pid<0){
		_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Unable to fork");
		/* Restore signal mask */
		sigprocmask(SIG_SETMASK, &oldmask, NULL);
		return 0;
	} else if(pid==0) {
		/* ======== Child ========= */
		int i;
		/* Order of setregid and setreuid is important, first group, then user */
		if (setregid(options.gid, options.gid)==0 && setreuid(options.uid, options.uid)==0){
			struct stat st;
			char *arge[4];
			va_list env;
			char *key;
			char *value;
			if (stat(options.cmd, &st) || !(S_ISREG(st.st_mode)) || !(st.st_mode&(S_IXUSR|S_IXGRP|S_IXOTH))) {
				_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "%s does not exist or is not executable", options.cmd);
				/* If the arguments supplied us with a script to run and it is not runable, return
				 * non-zero as this is an error.  Otherwise, return zero since it simply means the
				 * default doesn't exist, ie it is unconfigured and that is okay in our book */
				exit(options.cmdoverride);
			}
			/* Set up environment */
			va_start(env, numenv);
			for(i=0; i<numenv; i++){
				key   = va_arg(env, char*);
				value = va_arg(env, char*);
				if(value != NULL){
					setenv(key, value, 0);
				}
			}
			va_end(env);

			arge[0] = options.cmd;
			arge[1] = options.realuser;
			arge[2] = options.service; /* Pass service as second argument (thanks Hanno!) */
			arge[3] = '\0';
			execvp(arge[0], arge);
			/* If we reach this, execvp failed.  Exit with an error */
			_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "execvp error: %d", errno);
			exit(1);
		} else {
			_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Unable to run as uid %d", options.uid);
		}
		/* ============= Child Ends ======== */
	}
	/* if pid <=0, we would not be here, so pid>0, which means we are the parent */
	/* wait for the child to return */
	while(waitpid(pid, &retcode, 0)==-1){
		if(errno != EINTR){
			_pam_log(LOG_ERR, LOCAL_LOG_PREFIX "waitpid error: %d", errno);
			/* Restore signal mask */
			sigprocmask(SIG_SETMASK, &oldmask, NULL);
			return 0;
		}
	}
	if (!WIFEXITED(retcode)) {/* terminated abnormally */
		_pam_log(LOG_ERR,LOCAL_LOG_PREFIX "Executing file: \"%s\" as uid %d returned abnormally", options.cmd, options.uid);
		/* Restore signal mask */
		sigprocmask(SIG_SETMASK, &oldmask, NULL);
		return 0;
	}
	if(WEXITSTATUS(retcode)!=0){ /* Terminated normally, but non-zero */
		_pam_log(LOG_ERR,LOCAL_LOG_PREFIX "Executing file: \"%s\" as uid %d returned %d", options.cmd, options.uid, WEXITSTATUS(retcode));
		/* Restore signal mask */
		sigprocmask(SIG_SETMASK, &oldmask, NULL);
		return 0;
	}

	/* SUCCESS */
	/* Restore signal mask */
	sigprocmask(SIG_SETMASK, &oldmask, NULL);

	return 1;
}

/* Session management stuff */
PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	struct pamscript_options options;
	if(parse_options(pamh, argc, argv, "onsessionopen", &options) && session_script(options, 0)){
		return PAM_SUCCESS;
	}
	return PAM_SESSION_ERR;
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	struct pamscript_options options;
	if(parse_options(pamh, argc, argv, "onsessionclose", &options) && session_script(options, 0)){
		return PAM_SUCCESS;
	}
	return PAM_SESSION_ERR;
}

/* Authentication stuff */
PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	struct pamscript_options options;
	char *password;
	char *krb5ccname;
	if(!parse_options(pamh, argc, argv, "onauth", &options)){
		return PAM_AUTHINFO_UNAVAIL;
	}

	password = NULL;
	krb5ccname = NULL;;

	if(options.expose & EXPOSE_AUTHTOK){
		/* Find password */
		pam_get_item(pamh, PAM_AUTHTOK, (const void **)&password);
		/* If none, prompt user for one */
		if(password == NULL){
			int rc;
			rc = _get_authtok(pamh);
			if(rc != PAM_SUCCESS){
				return rc;
			}
			pam_get_item(pamh, PAM_AUTHTOK, (const void **)&password);
		}
	}

	if(options.expose & EXPOSE_KRB5CCNAME){
		krb5ccname = (char*)pam_getenv(pamh, "KRB5CCNAME");
	}

	if(session_script(options, 2, "PAM_AUTHTOK", password, "KRB5CCNAME", krb5ccname)){
		return PAM_SUCCESS;
	}

	return PAM_AUTH_ERR;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	/* No additional credentials to set */
	return PAM_IGNORE;
}
